﻿#define _CRT_SECURE_NO_WARNINGS /* prevents some Visual Studio warnings */

#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <windows.h>
typedef wchar_t UNICHAR;
#define UNICHAR_FOPEN(a, b) _wfopen(a, L ## b)
#define UNICHAR_STRCPY(a, b) wcscpy(a, b)
#define UNICHAR_STRCAT(a, b) wcscat(a, L ## b)
#define UNICHAR_STRLEN(a) wcslen(a)
#else
typedef char UNICHAR;
#define UNICHAR_FOPEN(a, b) fopen(a, b)
#define UNICHAR_STRCPY(a, b) strcpy(a, b)
#define UNICHAR_STRCAT(a, b) strcat(a, b)
#define UNICHAR_STRLEN(a) strlen(a)
#defi
#endif

#define SWAP16(x) \
( \
	(((uint16_t)((x) & 0x00FF)) << 8) | \
	(((uint16_t)((x) & 0xFF00)) >> 8)   \
)

#define SWAP32(x) \
( \
	(((uint32_t)((x) & 0x000000FF)) << 24) | \
	(((uint32_t)((x) & 0x0000FF00)) <<  8) | \
	(((uint32_t)((x) & 0x00FF0000)) >>  8) | \
	(((uint32_t)((x) & 0xFF000000)) >> 24)   \
)

static bool bigEndian;
static bool read8(FILE *f, uint8_t *out);
static bool readBE16(FILE *f, uint16_t *out);
static bool readBE32(FILE *f, uint32_t *out);
static bool createBak(UNICHAR *filename, FILE *f);

int main(int argc, char* argv[])
{
	FILE *f = NULL;
	uint8_t *dat = NULL;
#ifdef WIN32
	UNICHAR **argvw = NULL;
	bool didAllocateArgvW = false;
#endif

	// detect endianness
	volatile uint32_t endiantest32 = 0xFF;
	bigEndian = (*(uint8_t *)&endiantest32 != 0xFF);

	printf("diskcopystrip v0.4 - by 8bitbubsy, 2023\n");
	printf("=======================================\n\n");

	if (argc != 2)
	{
		printf("Usage: diskcopystrip <filename.image>\n");
#ifdef _WIN32
		printf("       (or drag and drop the file onto this executable)\n");
#endif
		printf("\n");
		printf("Strips Disk Copy 4.2 data from Macintosh floppy disk images.\n");
		printf("Note: Does not support compressed disk images!\n");
		printf("\n");
#ifdef _WIN32
		system("PAUSE");
#endif
		return -1;
	}

	UNICHAR *filename;
#ifdef _WIN32
	/* Read input filename in widechar type (Windows is the ultimate portability killer!)
	** (yes, I know I can just do int wmain(...), but MinGW-W32 doesn't like this)
	*/
	int32_t numArgs;
	argvw = CommandLineToArgvW(GetCommandLineW(), &numArgs);
	if (argvw == NULL || numArgs != 2)
	{
		printf("ERROR: Couldn't read command line argument!\n");
		goto Error;
	}
	didAllocateArgvW = true;
	filename = argvw[1];
#else
	filename = argv[1];
#endif

	f = UNICHAR_FOPEN(filename, "rb");
	if (f == NULL)
	{
		printf("ERROR: Can't open input file for reading!\n");
		goto Error;
	}

	// check disk type
	fseek(f, 0x50, SEEK_SET);
	uint8_t diskType;
	if (!read8(f, &diskType)) goto IOError;
	if (diskType > 3) goto Invalid;

	// check magic ID
	fseek(f, 0x52, SEEK_SET);
	uint16_t magicID;
	if (!readBE16(f, &magicID)) goto IOError;
	if (magicID != 0x0100) goto Invalid;

	// check data size
	fseek(f, 0x40, SEEK_SET);
	uint32_t dataSize;
	if (!readBE32(f, &dataSize)) goto IOError;
	if (dataSize < 409600 || dataSize > 1474560) goto Invalid;

	// get tag size
	fseek(f, 0x44, SEEK_SET);
	bool hasTagData;
	uint32_t tagSize;
	if (!readBE32(f, &tagSize)) goto IOError;
	hasTagData = (tagSize > 0);

	// OK. We're good to go...

	printf("Creating backup of disk image... ");
	if (!createBak(filename, f))
	{
		printf("FAILED! Exiting...\n");
		goto Error;
	}
	printf("OK.\n");

	// write data now

	dat = (uint8_t *)malloc(dataSize);
	if (dat == NULL)
	{
		printf("ERROR: Out of memory!\n");
		goto Error;
	}

	fseek(f, 0x54, SEEK_SET);
	if (fread(dat, 1, dataSize, f) != dataSize) goto IOError;
	fclose(f);

	f = UNICHAR_FOPEN(filename, "wb");
	if (f == NULL)
	{
		printf("ERROR: Can't open output file for writing!\n");
		goto Error;
	}

	if (fwrite(dat, 1, dataSize, f) != dataSize) goto IOError;
	fclose(f);
	free(dat);

	if (hasTagData)
		printf("Disk Copy header and tag data were successfully stripped.\n");
	else
		printf("Disk Copy header was successfully stripped.\n");

#ifdef _WIN32
	free(*argvw);
#endif
	return 0;

Invalid:
	printf("ERROR: This is not a Disk Copy image!\n");
	goto Error;
IOError:
	printf("ERROR: General I/O error during reading/writing!\n");
Error:

	if (f != NULL) fclose(f);
	if (dat != NULL) free(dat);

#ifdef _WIN32
	if (didAllocateArgvW) free(*argvw);
	printf("\n");
	system("PAUSE");
#endif

	return 1;
}

static bool read8(FILE *f, uint8_t *out)
{
	uint8_t data8;
	if (fread(&data8, 1, 1, f) != 1)
		return false;

	*out = data8;
	return true;
}

static bool readBE16(FILE *f, uint16_t *out)
{
	uint16_t data16;
	if (fread(&data16, 2, 1, f) != 1)
		return false;

	if (!bigEndian)
		data16 = SWAP16(data16);

	*out = data16;
	return true;
}

static bool readBE32(FILE *f, uint32_t *out)
{
	uint32_t data32;
	if (fread(&data32, 4, 1, f) != 1)
		return false;

	if (!bigEndian)
		data32 = SWAP32(data32);

	*out = data32;
	return true;
}

static bool createBak(UNICHAR *filename, FILE *f)
{
	uint8_t buffer[32768];
	FILE *fBak = NULL;
	UNICHAR *bakFilename = NULL;

	if (filename == NULL || f == NULL)
		return false;

	bakFilename = (UNICHAR *)malloc((UNICHAR_STRLEN(filename) + 1 + 3 + 1) * sizeof (UNICHAR));
	if (bakFilename == NULL)
		goto Error;

	UNICHAR_STRCPY(bakFilename, filename);
	UNICHAR_STRCAT(bakFilename, ".bak");

	fBak = UNICHAR_FOPEN(bakFilename, "wb");
	if (fBak == NULL)
		goto Error;

	const uint32_t oldPos = (uint32_t)ftell(f);

	fseek(f, 0, SEEK_END);
	const uint32_t filesize = (uint32_t)ftell(f);
	fseek(f, 0, SEEK_SET);
	
	uint32_t bytesLeft = filesize;
	while (bytesLeft > 0)
	{
		uint32_t bytesTodo = sizeof (buffer);
		if (bytesTodo > bytesLeft)
			bytesTodo = bytesLeft;
			
		if ( fread(buffer, 1, bytesTodo, f)    != bytesTodo) goto Error;
		if (fwrite(buffer, 1, bytesTodo, fBak) != bytesTodo) goto Error;
		
		bytesLeft -= bytesTodo;
	}
	
	fclose(fBak);
	free(bakFilename);

	fseek(f, oldPos, SEEK_SET);
	return true;
	
Error:
	if (fBak != NULL) fclose(fBak);
	if (bakFilename != NULL) free(bakFilename);
	
	fseek(f, oldPos, SEEK_SET);
	return false;
}
